/*--------------------------------------------------*
 |xmodemlights.c1.0                                 |
 |                                                  |
 | X Modem Lights display                           |
 |                                                  |
 | Graphical display of Linux PPP connection status |
 | indicates existence of connection, and if up,    |
 | continually monitors and the send/receive status |
 | of the link.                                     |
 |                                                  |
 | intended for use with AfterStep's Wharf.         |
 |                                                  |
 | requires Xlib, XPM library, and shape extension. |
 |                                                  |
 |david w. Tharp                             13MAR98|
 *--------------------------------------------------*/

/*
 *  THIS FILE IS VERY POORLY COMMENTED (SORRY) 
 *  BUT IT'S SO RIDICULOUSLY SIMPLE THAT IT SHOULDN'T 
 *  BE HARD TO FOLLOW ANYWAY.  I MAY COMMENT IT SOMEDAY...
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/xpm.h>
#include <X11/Xmu/Drawing.h>
#include <X11/extensions/shape.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "xpm/disconnected.xpm"
#include "xpm/connected.xpm"
#include "xpm/litup.xpm"
#include "xpm/lightregions"

/* some handy macros */

#define XSCM(img) XShapeCombineMask(display, window, ShapeBounding, \
	0, 0, img, ShapeSet);

#define XCPFD(xpm,img,shp) XpmCreatePixmapFromData(display, window, \
	xpm, img, shp, (XpmAttributes *) NULL);

#define XCWF XClearWindow(display, window);\
	XFlush(display);


/* global variables */
Display *             display;
int                   screen, depth;
Window                window;
XSetWindowAttributes  winattr;
unsigned long         winmask;
XSizeHints            hints;
char                  winname[] = "xmodemlights";
Pixmap                disconnectedshape, connectedshape, disconnected, connected, litup;
XGCValues             xgcvals;
GC                    gc;


/* functions */
int pppup();
void pppwait();
void pppmon();

/******************* main *******************
 *                                          */
int main(int argc, char ** argv) {
	
	display = XOpenDisplay("");
	if(display == NULL) {
		fprintf(stderr, "\n%s: %s\n\n", argv[0],"Unable to connect to X Server.");
		exit(1);
	}

	screen = DefaultScreen(display);
	depth = DefaultDepth(display, screen);
	
	winmask = CWEventMask;
	
	winattr.event_mask = StructureNotifyMask | SubstructureNotifyMask | ResizeRedirectMask ;
	
	window = XCreateWindow(display, RootWindow(display, screen),
		0, 0, 48, 48, 0, depth, InputOutput, CopyFromParent, winmask, &winattr);
	
	XStoreName(display, window, winname);
	
	hints.flags = PSize;
	hints.width = 48;
	hints.height = 48;
	XSetNormalHints(display, window, &hints);
	
	/* create the pixmaps from the xpm data */
	XCPFD(disconnected_xpm, &disconnected, &disconnectedshape);
	XCPFD(connected_xpm, &connected, &connectedshape);
	XCPFD(litup_xpm, &litup, NULL);
	
	xgcvals.graphics_exposures = False;
	gc = XCreateGC(display, window, GCGraphicsExposures, &xgcvals);
	
	XMapWindow(display, window);
	
	while(1) {
		if(pppup()) {
			pppmon(); /* monitor link 'til it dies */
		} else {
			pppwait(); /* wait for the link to come up */
		}
	}
}
/*                                          *
 **************** end of main ***************/


int pppup() {
	static FILE * pppfile;

	if((pppfile = fopen("/var/run/ppp0.pid", "r")) != NULL) {
		fclose(pppfile);
		return 1;
	} else {
		return 0;
	}
}


/*
 * wait until the ppp0 link comes up, and return.
 */
void pppwait() {
	int accum = 0, stilldown = 1;
	XEvent event;

	XSCM(disconnectedshape);
	XSetWindowBackgroundPixmap(display, window, disconnected);
	XCWF;

	while(stilldown) {
		while(XPending(display)) {
			XCheckMaskEvent(display, StructureNotifyMask, &event);
			switch(event.type) {
				case 17:
					exit(1);
			}
		}
		if(++accum >= 20) {
			stilldown = !pppup();
			accum = 0;
		}
		
		XNoOp(display);
		usleep(100000);
	}
	return;
}


/*
 * retrieve the number of packets received/sent from /proc/net/dev
 */
void getsr(long int * r, long int * s) {
	static char str[128];
	static char dev[32];
	static FILE * procnetdev;
	long int j1, j2, j3, j4, j5, j6, j7, j8, j9;
	
	procnetdev = fopen("/proc/net/dev", "r");
	while(!strstr(fgets(str, 128, procnetdev), "ppp0")) {}
	fclose(procnetdev);

	sscanf(str, "%s %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld\n",
		&dev, r, &j1, &j2, &j3, &j4, s, &j5, &j6, &j7, &j8, &j9);
}


/* 
 * monitor the ppp0 link while it's alive, and return when it dies.
 */
void pppmon() {
	long int sent, rcvd, lsent, lrcvd;
	unsigned char stillup = 1, accum = 0;
	XEvent event;
	Bool flush;

	XSCM(connectedshape);
	XSetWindowBackgroundPixmap(display, window, connected);
	XCWF;
	
	getsr(&rcvd, &sent);

	while(stillup) {
		flush = False;
		while(XPending(display)) {
			XCheckMaskEvent(display, StructureNotifyMask, &event);
			switch(event.type) {
				case 17:
					exit(1);
			}
		}
		XNoOp(display);
		lsent = sent;
		lrcvd = rcvd;
		getsr(&rcvd, &sent);
		if(accum == 0) {
			XClearWindow(display, window);
			flush = True;
		}
		if(++accum >= 20) {
			stillup = pppup();
			accum = 1;
		}
		if(sent != lsent) {
			XCopyArea(display, litup, window, gc, send_x, send_y, send_w, send_h, send_x, send_y);
			accum = 0;
			flush = True;
		}
		if(rcvd != lrcvd) {
			XCopyArea(display, litup, window, gc, receive_x, receive_y, receive_w, receive_h, receive_x, receive_y);
			accum = 0;
			flush = True;
		}
		if(flush) XFlush(display);
		usleep(100000);
	}
	return;
}
